#include "dbms.h"
#include "config.h"

#include <db.h>

#include <iostream>
#include <sstream>
#include <cstdlib>
#include <math.h>
#include <string.h>

#define BDB_FOLDER "/tmp/BDB/"
#define BDB_PARTITION false

#define ID(ana_key) (*((int*)((ana_key).data)))
#define DATE(ana_key) (*((uint64_t *)(((unsigned char *)((ana_key).data))+sizeof(int))))
#define VALUEA(ana_val) (*((float *)((ana_val).data)))
#define QUALITYA(ana_val) (*(((unsigned char *)((ana_val).data))+sizeof(float)))
#define VALUEB(bool_val) (*((bool *)((bool_val).data)))
#define QUALITYB(bool_val) (*(((unsigned char *)((bool_val).data))+sizeof(bool)))

using namespace std;

inline unsigned short endian_swap16(unsigned short x)
{
  return((x>>8) |
      (x<<8));
}

inline uint32_t endian_swap32(uint32_t x)
{
  return((x>>24) |
      ((x<<8) & 0x00FF0000) |
      ((x>>8) & 0x0000FF00) |
      (x<<24));
}

inline uint64_t endian_swap64(uint64_t x)
{
  return((x>>56) | 
      ((x<<40) & 0x00FF000000000000LL) |
      ((x<<24) & 0x0000FF0000000000LL) |
      ((x<<8)  & 0x000000FF00000000LL) |
      ((x>>8)  & 0x00000000FF000000LL) |
      ((x>>24) & 0x0000000000FF0000LL) |
      ((x>>40) & 0x000000000000FF00LL) |
      (x<<56));
}

static DBT ana_key;
static DBT ana_val;
static DB  * ana_db;
static DBC * ana_dbcp[CFG_ANA_TAGS];
static DBC * ana_dbc;
static DBT bool_key;
static DBT bool_val;
static DB  * bool_db;
static DBC * bool_dbcp[CFG_TOR_TAGS];
static DBC * bool_dbc;

#define BDB_PAGESIZE 4096
#define BDB_CACHESIZE 100*1024*1024
#define BDB_LOGREGIONSIZE 10*1024*1024
#define BDB_LKMAXOBJS 10000

const char * dbms::name()
{
  return "Berkeley DB";
}

const char * dbms::version()
{
  return "5.3.21";
}

const char * dbms::filesystem()
{
  return "ext4";
}

void dbms::print_options()
{
  cout << "pagesize: " << BDB_PAGESIZE << endl;
  cout << "cachesize: " << BDB_CACHESIZE << endl;
  cout << "lg_regionmax: " << BDB_LOGREGIONSIZE << endl;
  cout << "lk_max_objects: " << BDB_LKMAXOBJS << endl;
  cout << "partitions: ";
  if(BDB_PARTITION)
  {
  cout << CFG_ANA_TAGS << " (ana) + "
       << CFG_TOR_TAGS << " (tor)" << endl;
  }
  else
  {
    cout << "none" << endl;
  }
  cout << "data_path: " << BDB_FOLDER << endl;
  cout << "DB_TXN_NOSYNC" << endl;
  cout << "DB_READ_UNCOMMITTED" << endl;
  cout << "DB_CURSOR_BULK" << endl;
}

#define PNB 300

static u_int32_t part_fct(DB *db, DBT *key)
{
  return endian_swap32(ID(*key));
}

dbms::dbms()
{
  int ret;
  db_create(&ana_db, NULL, 0);

  DB_ENV * ana_env = ana_db->get_env(ana_db);

  ret = ana_env->set_cachesize(ana_env, 0, BDB_CACHESIZE, 1);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_cachesize error : %s (%i)\n",
        db_strerror(ret), ret);
  }
  
  ret = ana_env->set_lg_regionmax(ana_env, BDB_LOGREGIONSIZE);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_lg_regionmax error : %s (%i)\n",
        db_strerror(ret), ret);
  }

  ret = ana_env->set_lk_max_objects(ana_env, BDB_LKMAXOBJS);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_lk_max_objects error : %s (%i)\n",
        db_strerror(ret), ret);
  }

  ret = ana_env->set_flags(ana_env, DB_TXN_NOSYNC, 1);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_flags error : %s (%i)\n",
        db_strerror(ret), ret);
  }
  
  ret = ana_db->set_pagesize(ana_db, BDB_PAGESIZE);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_pagesize error : %s (%i)\n",
        db_strerror(ret), ret);
  }

  if(BDB_PARTITION && CFG_ANA_TAGS > 1)
  {
    ret = ana_db->set_partition(ana_db, CFG_ANA_TAGS, NULL, &(part_fct));
    if(ret != 0)
    {
      fprintf(stderr, "bdb set_partition error : %s (%i)\n",
          db_strerror(ret), ret);
    }
  }

  uint32_t flags = DB_READ_UNCOMMITTED;
  if(!CFG_SELECT_ONLY)
  {
    flags = flags | DB_CREATE | DB_TRUNCATE; //DB_THREAD
  }

  ret = ana_db->open(ana_db, NULL, BDB_FOLDER "bdb_ana", NULL,
      DB_BTREE, flags, 0664);
  if(ret != 0)
  {
    fprintf(stderr, "bdb_ana open error : %s (%i)\n",
        db_strerror(ret), ret);
    ana_db->close(ana_db, 0);
    exit(-1);
  }

  memset(&ana_key, 0, sizeof(DBT));
  memset(&ana_val, 0, sizeof(DBT));
  ana_key.size = sizeof(int) + sizeof(uint64_t);
  ana_key.data = (unsigned char *)malloc(ana_key.size);
  ana_val.size = sizeof(float)+ sizeof(unsigned char);
  ana_val.data = (unsigned char *)malloc(ana_val.size);

  ana_db->cursor(ana_db, NULL, &ana_dbc, 0);

  for(int i=0; i < CFG_ANA_TAGS; i++)
  {
    ret = ana_db->cursor(ana_db, NULL, &(ana_dbcp[i]), DB_CURSOR_BULK);
    if(ret != 0)
    {
      fprintf(stderr, "cursor open error %s (%i)\n", db_strerror(ret), ret);
    }
  }
 
  db_create(&bool_db, NULL, 0);
  
  DB_ENV * bool_env = bool_db->get_env(bool_db);
  
  ret = bool_db->set_pagesize(bool_db, BDB_PAGESIZE);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_pagesize error : %s (%i)\n",
        db_strerror(ret), ret);
  }
  
  ret = bool_env->set_cachesize(bool_env, 0, BDB_CACHESIZE, 1);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_cachesize error : %s (%i)\n",
        db_strerror(ret), ret);
  }

  ret = bool_env->set_lg_regionmax(bool_env, BDB_LOGREGIONSIZE);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_lg_regionmax error : %s (%i)\n",
        db_strerror(ret), ret);
  }

  ret = bool_env->set_flags(bool_env, DB_TXN_NOSYNC, 1);
  if(ret != 0)
  {
    fprintf(stderr, "bdb set_flags error : %s (%i)\n",
        db_strerror(ret), ret);
  }
  
  if(BDB_PARTITION && CFG_TOR_TAGS > 1)
  {
    ret = bool_db->set_partition(bool_db, CFG_TOR_TAGS, NULL, &(part_fct));
    if(ret != 0)
    {
      fprintf(stderr, "bdb set_partition error : %s (%i)\n",
          db_strerror(ret), ret);
    }
  }

  ret = bool_db->open(bool_db, NULL, BDB_FOLDER "bdb_tor", NULL,
      DB_BTREE, flags, 0664);
  if(ret != 0)
  {
    fprintf(stderr, "bdb_tor open error : %s (%i)\n",
        db_strerror(ret), ret);
    bool_db->close(bool_db, 0);
    exit(-1);
  }

  memset(&bool_key, 0, sizeof(DBT));
  memset(&bool_val, 0, sizeof(DBT));
  bool_key.size = sizeof(int) + sizeof(uint64_t);
  bool_key.data = (unsigned char *)malloc(bool_key.size);
  bool_val.size = sizeof(bool)+ sizeof(unsigned char);
  bool_val.data = (unsigned char *)malloc(bool_val.size);

  bool_db->cursor(bool_db, NULL, &bool_dbc, 0);

  for(int i=0; i < CFG_TOR_TAGS; i++)
  {
    ret = bool_db->cursor(bool_db, NULL, &(bool_dbcp[i]), DB_CURSOR_BULK);
    if(ret != 0)
    {
      fprintf(stderr, "cursor open error %s (%i)\n", db_strerror(ret), ret);
    }
  }
}

dbms::~dbms()
{
  ana_dbc->close(ana_dbc);
  ana_db->close(ana_db, 0);
  bool_dbc->close(bool_dbc);
  bool_db->close(bool_db, 0);
}

typedef struct ana_st
{
  char * label;
  uint32_t period;
  float threshold1;
  float threshold2;
} ana_t;

static ana_t ana_tbl[CFG_ANA_TAGS];

void dbms::insertAna(uint32_t anaid, char * label, uint32_t creationdate, uint32_t * destructiondate, uint32_t unit, uint32_t period, float threshold1, float threshold2)
{
  ana_tbl[anaid - 1].label = (char *)malloc(strlen(label) + 1);
  strcpy(ana_tbl[anaid - 1].label, label);
  ana_tbl[anaid - 1].period = period;
  ana_tbl[anaid - 1].threshold1 = threshold1;
  ana_tbl[anaid - 1].threshold2 = threshold2;
}

void dbms::insertBool(uint32_t boolid, char * label, uint32_t creationdate, uint32_t * destructiondate, char * label0, char * label1)
{}

void dbms::flushR0_1()
{
  ana_db->sync(ana_db, 0);
}

void dbms::flushR0_2()
{
  bool_db->sync(bool_db, 0);
}

void dbms::queryR0_1(uint32_t anaid, uint64_t date, float ana_value, unsigned char quality)
{
  if(!CFG_SELECT_ONLY)
  {
    ID(ana_key) = endian_swap32(anaid);
    DATE(ana_key) = endian_swap64(date);
    VALUEA(ana_val) = ana_value;
    QUALITYA(ana_val) = quality;

    int ret = ana_dbcp[anaid - 1]->put(ana_dbcp[anaid - 1], &ana_key, &ana_val, DB_KEYLAST);
    if(ret != 0)
    {
      fprintf(stderr, "put error : %s(%i)\n", db_strerror(ret), ret);
      exit(-1);
    }
  }
}

void dbms::queryR0_2(uint32_t boolid, uint64_t date,
    bool bool_value, unsigned char quality)
{
  if(!CFG_SELECT_ONLY)
  {
    ID(bool_key) = endian_swap32(boolid);
    DATE(bool_key) = endian_swap64(date);
    VALUEB(bool_val) = bool_value;
    QUALITYB(bool_val) = quality;

    int ret = bool_dbcp[boolid - 1]->put(bool_dbcp[boolid - 1], &bool_key, &bool_val, DB_KEYLAST);
    if(ret != 0)
    {
      fprintf(stderr, "put error : %s(%i)\n", db_strerror(ret), ret);
      exit(-1);
    }
  }

}

void dbms::queryR1_1(float Value, uint32_t AnaId, uint64_t Date)
{
  if(!CFG_SELECT_ONLY)
  {
    ID(ana_key) = endian_swap32(AnaId);
    DATE(ana_key) = endian_swap64(Date);

    int ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET);
    if(ret != 0)
    {
      fprintf(stderr, "R1.1 get error : %s(%i)\n", db_strerror(ret), ret);
      fprintf(stderr, "position (%i, %lld)\n", AnaId,
          (unsigned long long)Date);
      exit(-1);
    }

    QUALITYA(ana_val) = QUALITYA(ana_val) | 128;
    VALUEA(ana_val) = Value;

    ana_dbc->put(ana_dbc, &ana_key, &ana_val, DB_CURRENT);
  }
}

void dbms::queryR1_2(bool Value, uint32_t BoolId,
    uint64_t Date)
{
  if(!CFG_SELECT_ONLY)
  {
    ID(bool_key) = endian_swap32(BoolId);
    DATE(bool_key) = endian_swap64(Date);

    int ret = bool_dbc->get(bool_dbc, &bool_key, &bool_val, DB_SET);
    if(ret != 0)
    {
      fprintf(stderr, "R1.2 get error : %s(%i)\n", db_strerror(ret), ret);
      fprintf(stderr, "position (%i, %lld)\n", BoolId,
          (unsigned long long)Date);
      exit(-1);
    }

    QUALITYB(bool_val) = QUALITYB(bool_val) | 128;
    VALUEB(bool_val) = Value;

    bool_dbc->put(bool_dbc, &bool_key, &bool_val, DB_CURRENT);
  }
}

void dbms::queryR2_1(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)
    (uint32_t, uint64_t, float, unsigned char))
{
  ID(ana_key) = endian_swap32(AnaId);
  DATE(ana_key) = endian_swap64(start);

  int ret;

  ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  while(endian_swap32(ID(ana_key)) == AnaId && endian_swap64(DATE(ana_key)) < end)
  {
    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }

    callback(endian_swap32(ID(ana_key)), endian_swap64(DATE(ana_key)), VALUEA(ana_val), QUALITYA(ana_val));
  }
}

void dbms::queryR2_2(uint32_t BoolId, uint64_t start,
    uint64_t end, void (*callback)
    (uint32_t, uint64_t, bool, unsigned char))
{
  ID(bool_key) = endian_swap32(BoolId);
  DATE(bool_key) = endian_swap64(start);

  int ret;

  ret = bool_dbc->get(bool_dbc, &bool_key, &bool_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  while(endian_swap32(ID(bool_key)) == BoolId && endian_swap64(DATE(bool_key)) < end)
  {
    ret = bool_dbc->get(bool_dbc, &bool_key, &bool_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }

    callback(endian_swap32(ID(bool_key)), endian_swap64(DATE(bool_key)), VALUEB(bool_val), QUALITYB(bool_val));
  }
}

void dbms::queryR3_1(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(uint64_t))
{
  ID(ana_key) = endian_swap32(AnaId);
  DATE(ana_key) = endian_swap64(start);

  int ret;
  int count = 0;

  ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  while(endian_swap32(ID(ana_key)) == AnaId && endian_swap64(DATE(ana_key)) < end)
  {
    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }

    count++;
  }
  callback(count);
}

void dbms::queryR3_2(uint32_t BoolId, uint64_t start,
    uint64_t end, void (*callback)(uint64_t))
{
  ID(bool_key) = endian_swap32(BoolId);
  DATE(bool_key) = endian_swap64(start);

  int ret;
  int count = 0;

  ret = bool_dbc->get(bool_dbc, &bool_key, &bool_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  while(endian_swap32(ID(bool_key)) == BoolId && endian_swap64(DATE(bool_key)) < end)
  {
    ret = bool_dbc->get(bool_dbc, &bool_key, &bool_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }

    count++;
  }
  callback(count);
}

void dbms::queryR4(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(double))
{
  ID(ana_key) = endian_swap32(AnaId);
  DATE(ana_key) = endian_swap64(start);

  int ret;
  double sum = 0;

  ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  while(endian_swap32(ID(ana_key)) == AnaId && endian_swap64(DATE(ana_key)) < end)
  {
    sum+=VALUEA(ana_val);

    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }
  }
  callback(sum);
}

void dbms::queryR5(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(double))
{
  ID(ana_key) = endian_swap32(AnaId);
  DATE(ana_key) = endian_swap64(start);

  int ret;
  double sum = 0;
  long long count = 0;

  ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  while(endian_swap32(ID(ana_key)) == AnaId && endian_swap64(DATE(ana_key)) < end)
  {
    count++;
    sum+=VALUEA(ana_val);

    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }
  }
  callback(sum/count);
}

void dbms::queryR6(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(float, float))
{
  ID(ana_key) = endian_swap32(AnaId);
  DATE(ana_key) = endian_swap64(start);

  int ret;
  float min = NAN;
  float max = NAN;

  ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  if(endian_swap32(ID(ana_key)) == AnaId && endian_swap64(DATE(ana_key)) < end)
  {
    min = VALUEA(ana_val);
    max = VALUEA(ana_val);
  }

  while(endian_swap32(ID(ana_key)) == AnaId && endian_swap64(DATE(ana_key)) < end)
  {
    if(min > VALUEA(ana_val)) min = VALUEA(ana_val);
    if(max < VALUEA(ana_val)) max = VALUEA(ana_val);

    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }
  }
  callback(min, max);
}

void dbms::queryR7(uint32_t AnaId, uint64_t start,
    uint64_t end, void (*callback)(uint64_t, float))
{
  ID(ana_key) = endian_swap32(AnaId);
  DATE(ana_key) = endian_swap64(start);

  int ret;
  float threshold = ana_tbl[AnaId - 1].threshold2;

  ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  while(endian_swap32(ID(ana_key)) == AnaId && endian_swap64(DATE(ana_key)) < end)
  {
    if(VALUEA(ana_val) > threshold)
    {
      callback(endian_swap64(DATE(ana_key)), VALUEA(ana_val));
    }

    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }
  }
}

void dbms::queryR8(uint32_t AnaId, uint64_t start,
    uint64_t end, float Threshold, void (*callback)
    (uint64_t, float))
{
  ID(ana_key) = endian_swap32(AnaId);
  DATE(ana_key) = endian_swap64(start);

  int ret;

  ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
  if(ret != 0)
  {
    cerr << "get (SET_RANGE) error : " << ret << endl;
    exit(-1);
  }

  while(endian_swap32(ID(ana_key)) == AnaId && endian_swap64(DATE(ana_key)) < end)
  {
    if(VALUEA(ana_val) > Threshold)
    {
      callback(endian_swap64(DATE(ana_key)), VALUEA(ana_val));
    }

    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
    if(ret != 0)
    {
      cerr << "get (NEXT) error : " << ret << endl;
      exit(-1);
    }
  }
}

void dbms::queryR9(uint64_t start, uint64_t end,
    void (*callback)(const char *, uint64_t))
{
  int ret;
  uint64_t counts[CFG_ANA_TAGS];

  for(uint32_t i = 1; i <= CFG_ANA_TAGS; i++)
  {
    counts[i - 1] = 0;
    float threshold1 = ana_tbl[i - 1].threshold1;
    float threshold2 = ana_tbl[i - 1].threshold2;

    ID(ana_key) = endian_swap32(i);
    DATE(ana_key) = endian_swap64(start);

    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
    if(ret != 0)
    {
      cerr << "get (SET_RANGE) error : " << ret << endl;
      exit(-1);
    }

    while(endian_swap32(ID(ana_key)) == i && endian_swap64(DATE(ana_key)) < end)
    {
      if(VALUEA(ana_val) > threshold2 || VALUEA(ana_val) < threshold1)
      {
        counts[i - 1]++;
      }

      ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
      if(ret != 0)
      {
        cerr << "get (NEXT) error : " << ret << endl;
        exit(-1);
      }
    }
  }

  uint32_t max_id = 0;
  uint64_t max_count = 0;

  for(uint32_t i = 0; i < CFG_ANA_TAGS; i++)
  {
    if(counts[i] >= max_count)
    {
      max_count = counts[i];
      max_id = i;
    }
  }

  callback(ana_tbl[max_id].label, max_count);
}

void dbms::queryR10(uint64_t start, uint64_t end,
    void (*callback)(uint32_t, uint64_t))
{
  int ret;
  uint64_t counts[CFG_ANA_TAGS];

  for(uint32_t i = 1; i <= CFG_ANA_TAGS; i++)
  {
    counts[i - 1] = 0;
    uint32_t period = ana_tbl[i - 1].period * 1000;
    uint64_t prevdate;

    ID(ana_key) = endian_swap32(i);
    DATE(ana_key) = endian_swap64(start);

    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
    if(ret != 0)
    {
      cerr << "get (SET_RANGE) error : " << ret << endl;
      exit(-1);
    }

    prevdate = endian_swap64(DATE(ana_key));

    while(endian_swap32(ID(ana_key)) == i && prevdate < end)
    {
      ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_NEXT);
      if(ret != 0)
      {
        cerr << "get (NEXT) error : " << ret << endl;
        exit(-1);
      }

      if((endian_swap64(DATE(ana_key)) - prevdate) > period)
      {
        counts[i - 1]++;
      }
      prevdate = endian_swap64(DATE(ana_key));
    }
  }
  
  uint32_t max_id = 0;
  uint64_t max_count = 0;

  for(uint32_t i = 0; i < CFG_ANA_TAGS; i++)
  {
    if(counts[i] >= max_count)
    {
      max_count = counts[i];
      max_id = i;
    }
  }
  callback(max_id + 1, max_count);
}

void dbms::queryR11_1(void (*callback)(uint32_t, float))
{
  int ret;

  for(uint32_t i = 1; i <= CFG_ANA_TAGS; i++)
  {
    ID(ana_key) = endian_swap32(i+1);
    DATE(ana_key) = endian_swap64(0);

    ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_SET_RANGE);
    if(ret != 0 && ret != DB_NOTFOUND)
    {
      cerr << "get (SET_RANGE) error : " << ret << endl;
      exit(-1);
    }
    if(ret == DB_NOTFOUND)
    {
      ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_LAST);
      if(ret != 0)
      {
        cerr << "get (LAST) error : " << ret << endl;
        exit(-1);
      }
    }
    else
    {
      ret = ana_dbc->get(ana_dbc, &ana_key, &ana_val, DB_PREV);
      if(ret != 0)
      {
        cerr << "get (PREV) error : " << ret << endl;
        exit(-1);
      }
    }

    callback(i, VALUEA(ana_val));
  }
}

void dbms::queryR11_2(void (*callback)(uint32_t, bool))
{
  int ret;

  for(uint32_t i = 1; i <= CFG_TOR_TAGS; i++)
  {
    ID(bool_key) = endian_swap32(i+1);
    DATE(bool_key) = endian_swap64(0);

    ret = bool_dbc->get(bool_dbc, &bool_key, &bool_val, DB_SET_RANGE);
    if(ret != 0 && ret != DB_NOTFOUND)
    {
      cerr << "get (SET_RANGE) error : " << ret << endl;
      exit(-1);
    }
    if(ret == DB_NOTFOUND)
    {
      ret = bool_dbc->get(bool_dbc, &bool_key, &bool_val, DB_LAST);
      if(ret != 0)
      {
        cerr << "get (LAST) error : " << ret << endl;
        exit(-1);
      }
    }
    else
    {
      ret = bool_dbc->get(bool_dbc, &bool_key, &bool_val, DB_PREV);
      if(ret != 0)
      {
        cerr << "get (PREV) error : " << ret << endl;
        exit(-1);
      }
    }

    callback(i, VALUEB(bool_val));
  }
}

